@@ -16,7 +16,7 @@
 import asyncio
 import collections
 import copy
-import json
+import jsonpatch
 import kubernetes_asyncio as kubernetes
 import logging
 
@@ -35,10 +35,10 @@ LOG.setLevel(logging.INFO)
 # Prometheus metrics
 METRICS_KWARGS = dict(namespace="adaptdl", subsystem="sched")
 JOB_SUBMISSION_COUNT = Counter(
-        "job_submission_count", "Number of submitted jobs", **METRICS_KWARGS)
+    "job_submission_count", "Number of submitted jobs", **METRICS_KWARGS)
 JOB_COMPLETION_TIME = Summary(
-        "job_completion_time", "Duration of completed jobs",
-        labelnames=["status"], **METRICS_KWARGS)
+    "job_completion_time", "Duration of completed jobs",
+    labelnames=["status"], **METRICS_KWARGS)
 
 
 class AdaptDLController(object):
@@ -102,23 +102,24 @@ class AdaptDLController(object):
         current_ts = datetime.now(timezone.utc)
         job, pods = await self._get_job_and_pods(namespace, job_name)
         if job is not None:
-            job = await self._validate_job(job, pods, current_ts)
+            job = await self._validate_pods(job, pods, current_ts)
         if job is None:  # Not Found, presumably was deleted.
             await self._delete_pods(pods)
             return
+        # Use ChainMap to record updates to the job status fields.
+        job["status"] = collections.ChainMap({}, job.get("status", {}))
         # Get the current phase of the job, None if no phase was set.
         allocation = job["status"].get("allocation", [])
-        # Use ChainMap to record updates to the job status fields.
-        job["status"] = collections.ChainMap({}, job["status"])
-        phase = job["status"]["phase"]
+        phase = job["status"].setdefault("phase", "Pending")
         replicas = job["status"].get("replicas", 0)
-        if (completion_status := self._detect_completion(pods)):
+        preemptible = job["spec"].get("preemptible", True)
+        if (completion_status := self._detect_completion(pods, preemptible)):
             # Job is already completed.
             job["status"].update(completion_status)
             job["status"].setdefault("completionTimestamp", current_ts)
             job["status"]["allocation"] = allocation = []
             await self._delete_pods(  # Keep failed pods for debug purposes.
-                    [pod for pod in pods if pod.status.phase != "Failed"])
+                [pod for pod in pods if pod.status.phase != "Failed"])
         elif phase == "Pending":
             if allocation and not pods:
                 # Start the next group of pods.
@@ -204,26 +205,10 @@ class AdaptDLController(object):
             raise  # Unexpected error.
         return job, pods
 
-    async def _validate_job(self, job, pods, current_ts):
+    async def _validate_pods(self, job, pods, current_ts):
         namespace = job["metadata"]["namespace"]
         name = job["metadata"]["name"]
         patch_status = {}
-        if not job.get("status"):
-            # Assuming empty status section means this is a newly created job.
-            JOB_SUBMISSION_COUNT.inc()
-            patch_status = {"phase": "Pending"}
-            # Validate pod template using a dry run.
-            template = {
-                "metadata": {"name": name, "namespace": namespace},
-                "template": job["spec"]["template"],
-            }
-            try:
-                await self._core_api.create_namespaced_pod_template(
-                    namespace, template, dry_run="All")
-            except kubernetes.client.rest.ApiException as exc:
-                patch_status["phase"] = "Failed"
-                patch_status["reason"] = "Invalid"
-                patch_status["message"] = json.loads(exc.body).get("message")
         # Validate pods for job.
         group_list = []
         replicas_list = []
@@ -264,7 +249,7 @@ class AdaptDLController(object):
                                           {"status": patch_status})
         return job
 
-    def _detect_completion(self, pods):
+    def _detect_completion(self, pods, preemptible):
         if not pods:
             return {}
 
@@ -303,7 +288,8 @@ class AdaptDLController(object):
                 # we might be temporarily out of pods on this node
                 LOG.warning(f"Pod {pod.metadata.name} is {pod.status.reason} "
                             f"on {pod.spec.node_name}")
-            elif pod.metadata.deletion_timestamp is not None or any143(pod):
+            elif preemptible and (pod.metadata.deletion_timestamp is not None
+                                  or any143(pod)):
                 # This pod was intentionally terminated.
                 LOG.warning(f"Pod {pod.metadata.name} terminated")
             else:
@@ -416,9 +402,21 @@ class AdaptDLController(object):
                     "name": "NVIDIA_VISIBLE_DEVICES",
                     "value": "none",
                 })
+        pod = self._patch_pods_and_containers(pod)
         await self._core_api.create_namespaced_pod(
             job_metadata["namespace"], pod)
 
+    def _patch_pods_and_containers(self, pod):
+        pod_patch = config.get_job_patch_pods()
+        container_patch = config.get_job_patch_containers()
+        if pod_patch:
+            pod = jsonpatch.apply_patch(pod, pod_patch)
+        if container_patch:
+            for idx, container in enumerate(pod["spec"]["containers"]):
+                pod["spec"]["containers"][idx] = (
+                    jsonpatch.apply_patch(container, container_patch))
+        return pod
+
     def _get_pod_name(self, job_metadata, group, rank):
         job_name = job_metadata["name"]
         job_uid = job_metadata["uid"]
